#!/usr/bin/perl -w

use strict;
use ExtUtils::MakeMaker;
require 5.004;

my $SRCDIR = "ccp4-6.1.1/lib/DiffractionImage";
my $LIBDIR = "$SRCDIR/.libs";
my $DO_BUILD_LIB = 0;

if(exists($ENV{DIFFIMG_LIBDIR})){
	$LIBDIR = $ENV{DIFFIMG_LIBDIR};
	print "Use libDiffImage in $LIBDIR. libDiffImage will NOT be built.\n";
}else{
	$DO_BUILD_LIB = 1;
	print "DIFFIMG_LIBDIR is not set. libDiffImage will be built.\n";
}

# Prepare for building
print "Generating wrapper interface with SWIG...\n";
my @cmd = (qw/swig -c++ -perl -shadow -outdir lib -o DiffractionImagePerlInterface.cpp/,"$SRCDIR/DiffractionImageSwig.i");
print join(" ",@cmd) . "\n";
my $ret = system(@cmd);
if($ret == -1){ die "Swig has not been installed.\n" }
$ret && die "Swig failed ($ret) [$!]\n";

# configure DiffractionImage
my $SUBDIRS = [];
if($DO_BUILD_LIB){
	system("cd ccp4-6.1.1/lib/DiffractionImage; ./configure") && die "Configuring libDiffImage was failed.";
	$SUBDIRS = [$SRCDIR];
}

# Create Makefile
WriteMakefile(
    NAME              => 'DiffractionImage',
    INC               => "-I$SRCDIR",
    LIBS              => '-lstdc++',
    OBJECT            => "DiffractionImagePerlInterface\$(OBJ_EXT) $LIBDIR/libDiffImage\$(LIB_EXT) $LIBDIR/libDiffractionImage\$(LIB_EXT)",
    VERSION           => '0.1.1',
    PREREQ_PM         => {}, # e.g., Module::Name => 1.1
    ($] >= 5.005 ?     ## Add these new keywords supported since 5.005
      (AUTHOR         => 'Dobuo OKAZAKI, BioKids Project') : ()),
    clean             => {FILES => 'DiffractionImagePerlInterface.cpp lib/DiffractionImage.pm'},
	DIR               => $SUBDIRS,
);

1;

