// DiffractionImageCBF.cpp
// Author: F. Remacle
// 25 January 2005
// 
// The bits of the DiffractionImage library which pertain to Raxis images
// images.

#include "DiffractionImage.h"
#include "cbf.h"
#include "cbf_simple.h" 
#include "cbf_string.h"
#include "cbf_byte_offset.h"


using namespace std;

namespace DI 
	{
	void DiffractionImage::loadCBF()
		{
		cbf_handle cbfh2;
		FILE* imagefile2;
		int test;
		//read the header		
		loadCBFHeader();    
		allocateImage();
		imagefile2 = fopen(filename.c_str(),"rb");
		if(format.compare("MINI-CBF") == 0)
			{
			//throw(DiffractionImageException("Image Reading for Mini-CBF not supported yet"));
			cbf_file* ccbbff;
			cbf_make_file(&ccbbff,imagefile2);
			char* header_info= new char[256];
			int skip=0;
			int sign=0;
			int nelements=0;
			size_t nread=0;
			int padding=0;
			string byte_ord="";
			while(strstr(header_info,"--CIF-BINARY-FORMAT-SECTION--")==NULL)
				{
				fgets(header_info,256,imagefile2);
				skip+=strlen(header_info);
				}
			while(strcmp(header_info,"\n")!=0)
				{
				fgets(header_info,256,imagefile2);
				if(strstr(header_info,"signed")!=NULL)
					sign=1;
				if(strstr(header_info,"LITTLE_ENDIAN")!=NULL)
					byte_ord="LITTLE_ENDIAN";
				if(strstr(header_info,"BIG_ENDIAN")!=NULL)
					byte_ord="BIG_ENDIAN";
				if(strstr(header_info,"X-Binary-Number-of-Elements:")!=NULL)
					sscanf(header_info,"%*s %d",&nelements);
				if(strstr(header_info,"X-Binary-Size-Padding:")!=NULL)
					sscanf(header_info,"%*s %d",&padding);
				skip+=strlen(header_info);
				}
			fseek(imagefile2,0,SEEK_SET);
			cbf_set_fileposition (ccbbff, skip,SEEK_SET);
			int* tmpimg;
			tmpimg=new int[height*width];
			test= cbf_decompress_byte_offset ((void *)tmpimg,sizeof(int),sign,
                               			nelements,&nread,0,32,sign,ccbbff,0,
										byte_ord.c_str(),2,(size_t) width,
										(size_t) height,1,(size_t) padding);
			for(int i=0;i<height*width;i++)
				{
				if(tmpimg[i]<0)
					image[i]=0;
				else
					image[i]=tmpimg[i];
				}
				free(tmpimg);
			}
		else
			{
			test=cbf_make_handle(&cbfh2);
			test=cbf_read_file(cbfh2,imagefile2,MSG_DIGEST);
			if(width != height)
				{
				int* tmpimg;
				tmpimg=new int[height*width];
				test=cbf_get_image (cbfh2, 0,0,(void *) tmpimg, sizeof(int), 1, (size_t) width, (size_t) height);
				//test=cbf_get_image (cbfh2, 0,0,(void *) tmpimg, sizeof(long), 1, (size_t) height, (size_t) width);
				for(int i=0;i<height*width;i++)
					{
					if(tmpimg[i]<0)
						image[i]=0;
					else
						image[i]=tmpimg[i];
					}
				free(tmpimg);
				}
			else
				test=cbf_get_image (cbfh2, 0,0,(void *) image, sizeof(int), 0, (size_t) width, (size_t) height);
			}
		}
    
	void DiffractionImage::loadCBFHeader() 
		{
		cbf_handle cbfh;
		cbf_goniometer cbf_gon;
		cbf_detector cbfd;
		FILE* imagefile;
		double wave,dist,size[2],beamFast,BeamSlow,pxFast,pxSlow,sizeFast;
		double sizeSlow, oscS, oscIncr, second, exp, bX,bY,dumm1,dumm2;
		size_t myH, myW;
		char* mydate;
		char* topline;
		char* scanaxis;
		int year, month, day, hour, minute, timez;
		imagefile = fopen(filename.c_str(),"rb");
		topline=new char[256];
		fgets(topline,256,imagefile);
		fclose(imagefile);
		if(strstr(topline,"PILATUS")!=NULL)
			{
			float ang_incr;
			char hash;
			string headertmp;
			char *header_info;
			char* header_copy;
			char* value;
			char* value2;
			char* beamUnit;
			int endline;
			int ninfo=0;
			int nosc=0;
			format="MINI-CBF";
			manufact="DECTRIS";
			imagefile = fopen(filename.c_str(),"rb");
			header_info=new char[256];
			while((header_info[0]=='#' && header_info[1]!=' ') ||
				   header_info[0]!='#')
				fgets(header_info,256,imagefile);
			while(ninfo < 13)
				{
				value=new char[20];
				value2=new char[20];
				sscanf(header_info,"# %s",value);
				if(header_info[2]=='2')
					{
					sscanf(header_info,"# %s %s",value,value2);
					date=value;
					date.append(1,' ');
					date.append(value2);
					ninfo++;
					}
				if(cbf_cistrcmp(value,"Pixel_size")==0)
					{
					sscanf(header_info,"# %*s %f m x %f m",&pixelX,&pixelY);
					pixelX*=1000;
					pixelY*=1000;
					ninfo+=2;
					}
				if(cbf_cistrcmp(value,"Exposure_time")==0)
					{
					sscanf(header_info,"# %*s %f",&exposureTime);
					ninfo++;
					}
				if(cbf_cistrcmp(value,"Wavelength")==0)
					{
					sscanf(header_info,"# %*s %f",&wavelength);
					ninfo++;
					}
				if(cbf_cistrcmp(value,"Detector_distance")==0)
					{
					sscanf(header_info,"# %*s %f",&distance);
					distance*=1000;
					ninfo++;
					}
				if(cbf_cistrcmp(value,"Start_angle")==0)
					{
					sscanf(header_info,"# %*s %f",&oscStart);
					ninfo++;
					}
				if(cbf_cistrcmp(value,"Angle_increment")==0)
					{
					sscanf(header_info,"# %*s %f",&ang_incr);
					ninfo++;
					}
				if(cbf_cistrcmp(value,"Detector_2theta")==0)
					{
					sscanf(header_info,"# %*s %f",&twoTheta);
					ninfo++;
					}
				if(cbf_cistrcmp(value,"N_oscillations")==0)
					{
					sscanf(header_info,"# %*s %d",&nosc);
					ninfo++;
					}
				if(cbf_cistrcmp(value,"Oscillation_axis")==0)
					{
					sscanf(header_info,"# %*s %s",value);
					oscAxis=value;
					//not a good fudge, need to look that in details sometime
					if(oscAxis.compare("X,"))
						oscAxis="phi";
					ninfo++;
					}
				if(cbf_cistrcmp(value,"Beam_xy")==0)
					{
					beamUnit=new char[10];
					sscanf(header_info,"# %*s (%f, %f) %s",&beamX,&beamY,beamUnit);
					ninfo+=2;
					}
				fgets(header_info,256,imagefile);
				}
			oscEnd=oscStart+nosc*ang_incr;
			if(cbf_cistrcmp(beamUnit,"pixels")==0)
				{
				beamX*=pixelX;
				beamY*=pixelY;
				}
			else
				{
				if(cbf_cistrcmp(beamUnit,"mm")!=0)
					throw(DiffractionImageException("MiniCBF Beam centre unit is neither pixel nor mm"));
				}
			while(strstr(header_info,"X-Binary-Size-")==NULL)
				fgets(header_info,256,imagefile);
			ninfo=0;
			while(ninfo<2)
				{
				if(strstr(header_info,"Fastest-Dimension")!=NULL)
					{
					sscanf(header_info,"%*s %d",&height);
					ninfo++;
					}
				if(strstr(header_info,"Second-Dimension")!=NULL)
					{
					sscanf(header_info,"%*s %d",&width);
					ninfo++;
					}
				if(strstr(header_info,"First-Dimension")!=NULL)
					{
					sscanf(header_info,"%*s %d",&height);
					ninfo++;
					}
				if(strstr(header_info,"Slowest-Dimension")!=NULL)
					{
					sscanf(header_info,"%*s %d",&width);
					ninfo++;
					}
				fgets(header_info,256,imagefile);
				}
			}
		else
			{
			imagefile = fopen(filename.c_str(),"rb");
			cbf_make_handle(&cbfh);
			cbf_read_file(cbfh,imagefile,MSG_DIGEST);
			format="CBF";
			const char* detecttype;
			cbf_find_tag (cbfh,"_diffrn_detector.type");
			cbf_get_value(cbfh,&detecttype);
			if(strstr(detecttype,"ADSC")!=NULL || strstr(detecttype,"QUANTUM")!=NULL)
				manufact="ADSC";
			if(strstr(detecttype,"MAR")!=NULL)
				manufact="MAR";
			if(strstr(detecttype,"PILATUS")!=NULL || strstr(detecttype,"Pilatus")!=NULL)
				manufact="DECTRIS";
			if((strstr(detecttype,"RIGAKU")!=NULL || strstr(detecttype,"RAXIS")!=NULL) ||
			   (strstr(detecttype,"R-AXIS")!=NULL || strstr(detecttype,"SATURN")!=NULL))
				manufact="RIGAKU";
			cbf_get_wavelength(cbfh,&wave);
			wavelength=(float) wave;
			//cbf_get_image_size(cbfh,0,0,&myH,&myW);  
			cbf_get_image_size(cbfh,0,0,&myW,&myH);
			height=(int) myH;
			width=(int) myW;
			//this is just in case the image as been recorded as a single dimension
			//array in the CBF format. We then do not have any way to make a difference
			//between width and height so we assume the image is square.
			if(height==1)
				{
				width=(int) sqrt((double) width);
				height=width;
				}
			cbf_construct_detector (cbfh, &cbfd, 0); 
			cbf_get_beam_center (cbfd, &dumm1,&dumm2, &bY, &bX); 
			beamX=(float) bX;
			beamY=(float) bY;
			//Probably not the way it should be done but no other choice by now.
			if(beamX < 0)
				beamX=-beamX;
			if(beamY < 0)
				beamY=-beamY;
			cbf_get_detector_distance(cbfd,&dist);
			distance=(float) dist;
			cbf_get_pixel_size(cbfh,0,1,&pxFast);
			pixelX=(float) pxFast;
			cbf_get_pixel_size(cbfh,0,2,&pxSlow);
			pixelY=(float) pxSlow;
			
			cbf_get_axis_setting(cbfh,0, "GONIOMETER_PHI", &oscS, &oscIncr);
			oscStart=(float) oscS;
			oscEnd=(float) (oscS+oscIncr);
			if(oscStart==oscEnd)
				{
				cbf_get_axis_setting(cbfh,0, "GONIOMETER_OMEGA", &oscS, &oscIncr);
				oscStart=(float) oscS;
				oscEnd=(float) (oscS+oscIncr);
				if(oscStart==oscEnd)
					{
					cbf_get_axis_setting(cbfh,0, "GONIOMETER_KAPPA", &oscS, &oscIncr);
					oscStart=(float) oscS;
					oscEnd=(float) (oscS+oscIncr);
					if(oscStart==oscEnd)
						oscAxis="unknown";
					else
						oscAxis="kappa";
					}
				else
					oscAxis="omega";
				}
			else
				{
				oscAxis="phi";
				}
			//oscAxis=scanaxis;
			
			cbf_get_datestamp (cbfh, 0, &year, &month, &day, &hour, &minute, &second, &timez);
			mydate=new char[80];
			sprintf(mydate,"%d/%d/%d %d:%d:%f",day,month,year,hour,minute,second);
			date=string(mydate);
			cbf_get_integration_time(cbfh,0,&exp);
			exposureTime=(float) exp;
			twoTheta=0;
			}
		fclose(imagefile);
		free(topline);
		}

/*	void DiffractionImage::saveCBF()
		{ 
		}	*/

	};
