// DiffractionImageJpeg.cpp
// authors: G.Winter, F.Remacle
// maintained by F.Remacle
// 17th December 2003
// 
// The Jpeg (old school) handling routines for the DiffractionImage
// module. In addition, all of the utility code for handling the image
// is contained herein.
// 
// 
// $Id: DiffractionImageImgOperations.cpp,v 1.15 2008/05/07 11:00:51 fr45 Exp $

#include "DiffractionImage.h"

namespace DI 
	{
	// this is a utility function so is not contained within the class 
	// structure
	void enlargeImage(unsigned int * original,int width,int height,int zoom,unsigned int * copy)
		{
		// this is pretty easy
		int i, j, k, l;
		int position;
	
		if (original == NULL)
			throw (DiffractionImageException("original image NULL"));
		if (copy == NULL)
			throw (DiffractionImageException("copy image NULL"));
		if ((width <= 0) || (height <= 0))
			throw (DiffractionImageException("wrong size"));
	
		// just a hack to prevent funny things going on
		if (zoom < 1)
			zoom = 1;
		// check for special case zoom == 1 => duplicate
		if (zoom == 1)
			{
			for (i = 0; i < width * height; i++)
				copy[i] = original[i];
			return;
			}
		for (i = 0; i < height; i++)
			{
			for (j = 0; j < width; j++)
				{
				for (k = 0; k < zoom; k++)
					{
					for (l = 0; l < zoom; l++)
						{
						// simply copy one pixel over a larger area
						position = l + k * width * zoom + j * zoom + i * zoom * width * zoom ;
						copy[position] = original[i * width + j];
						}
					}
				}
			}
		}
    
	// see above statement
	void reduceImage(unsigned int * original,int width,int height,int zoom,unsigned int * copy)
		{
		int i, j, k, l,width2,height2;
		int position;
		int accumulator;    
		if (original == NULL)
			throw (DiffractionImageException("original image NULL"));
		if (copy == NULL)
			throw (DiffractionImageException("copy image NULL"));
		if ((width <= 0) || (height <= 0))
			throw (DiffractionImageException("wrong size"));
	
		// check the zoom value
		if (zoom < 1)
			zoom = 1;

		// check for special case zoom == 1 => duplicate
		if (zoom == 1)
			{
			for (i = 0; i < width * height; i++)
				copy[i] = original[i];
			return;
			}
	
		// reset the input values to allow for the scaling - the above code 
		// will now work except it is now backwards
		// these routines will really bash the memory - probably a good
		// benchmark for memory bandwidth these!
		width2 = width/zoom;
		height2 = height/zoom;
			
		for (j = 0; j < width2; j++)
			{
			for (i = 0; i < height2; i++)
				{
				// for this, apply the average value over zoom x zoom
				// pixels as the new pixel value
				accumulator = 0;
				for (k = 0; k < zoom; k++)
					{
					for (l = 0; l < zoom; l++)
						{
						position = l + k * width + j * zoom + (i * zoom) * width;
						accumulator += original[position];
						}
					}
				accumulator /= (zoom * zoom);
				copy[i * width2 + j] = accumulator;
				}
			}
		}
 
		
    // this will convert a 32 bit unsigned long image to an unsigned 
    // char (byte) representation. theree are two options for these -
    // optimise and not optimise. the optimised version will perform
    // some floating point calculations to make a nicer image, but will
    // probably be quite a bit slower. the results from both should be 
    // tested.
    void downSample(unsigned int * original,int width,int height,bool optimise,unsigned char * copy)
		{
		// check that the input is ok
		if (original == NULL)
			throw (DiffractionImageException("original image NULL"));
		if (copy == NULL)
			throw (DiffractionImageException("copy image NULL"));
		if ((width <= 0) || (height <= 0))
			throw (DiffractionImageException("wrong image size"));
		
		// if not optimise then determine the appropriate left shiting to
		// do for each element - keeping integer arithmatic should make
		// this nice and fast
#ifdef TEST_CODE
		if (optimise)
			cout << "downsample: optmisation on" << endl;
		else
			cout << "downsample: optimisation off" << endl;
#endif
		if (!optimise)
			{
			int i, left,value;
#ifdef TEST_CODE
			cout << sizeof(long long) << endl;
#endif
			// a long long should be big enough to store the total
			// 8 bytes -> 18446744073709551616
			long long total = 0;
			left = 1;
	    
			// work over the image and work out, to the nearest power of
			// 2, the maximum value found in the image
		    for (i = 0; i < width * height; i++)
				{
				if (original[i] > (2 << left))
					left = left ++;
				total += original[i];
				}
			total /= (width * height);

			// want the average to be about "half" the byte - but this
			// may want to be tweaked

			while (total < (2 << (left - 4)))
				left --;

#ifdef TEST_CODE
			assert(left <= 32);
			cout << "downsample: shifting " << left << " bits" << endl;
			cout << "average: " << total << endl;
#endif
			// recall that we want to shift from 32 to 8 bit
			// so the amount to shift by is whatever is greater
			// than 24
			if (left < 24)
				left = 0;
			else
				left -= 24;
#ifdef TEST_CODE
			cout << "downsample: shifting " << left << " bits" << endl;
#endif
			// go through and copy byte shifted values into the new array  
			for (i = 0; i < width * height; i++)
				{
				value = original[i] >> left;
				// check that the values will fit into 1 bytes - which 
				// we know they will by design
				
				if (value > 0xff)
					value = 0xff;
				
				else
					{
					if (value < 0)
						value = 0;
					}

				// perform the assignment - with the typing
				// to prevent complaints
				copy[i] = (unsigned char) value;
				}
			}
		else
			{
			// we do floating point calculations to get a nice spread 
			// of values over this image

			// these are declared to be double as the average pixel value
			// over the entire image will be gathered up - and we don't 
			// want rounding errors, do we? :o)
			double min, max, average, total, current;
			int value, i;
	    
			// determine the average pixel value - this could 
			// probably be applied over a smaller subset of the image to 
			// good effect.
			total = 0;
			for (i = 0; i < width * height; i++)
				{
				if (original[i] < 0x8000)
					total += original[i];
				else
					total += 0x8000;
				}
	    
			// this is found to give a nice image for X-ray 
			// diffraction images - a dynamic range of sixteen seems
			// to bring out both the diffraction and background nicely.
			average = total / (width * height);
			min = 0;
			max = average * 4.0;
#ifdef TEST_CODE
			cout << "downsample: min " << min << " max " << max << endl;
#endif
			// now rescale the original pixel values into the new ranges
			// and store as bytes
			for (i = 0; i < width * height; i++)
				{
				current = original[i];
				value = (int) (0xff * (current - min) / (max - min));
				// this checking will definately be needed, as there
				// will by design be a large number of pixels which 
				// exceed the maximum value
				if (value > 0xff)
					value = 0xff;
				else
					{
					if (value < 0)
						value = 0;
					}
				// assign finally
				copy[i] = (unsigned char) value;
				}
			}
		}
		
	// next a method to invert the image, since we will probably want that
	// x-ray diffraction images are normally displayed as a negative, 
	// with the areas of high x-ray intensity (i.e. spots) shown as black. 
	void invertImage(unsigned char * image,int width,int height)
		{
		// value checking
		if (image == NULL)
			throw (DiffractionImageException("NULL image"));
		if ((width <= 0) || (height <= 0))
			throw (DiffractionImageException("wrong size"));
		int i;
		// simply compute the additive inverse
		for (i = 0; i < width * height; i++)
			image[i] = 0xff - image[i];
		}

	//This one is needed to be able to get read from Tcl interface.
	char* DiffractionImage::getImage(double rate,int upleftX,int upleftY,int boxW,int boxH)
		{
		unsigned int* firstcopy;
		unsigned int* cropped;
		unsigned char* copy1;
		unsigned char* copy2;
		char* mytemplate;
		FILE* mytempfile;
		int filed;
		int test,k;
		int rateint,newwidth,newheight;
			
		if(rate < 1)
			{
			rateint=floor(1.0/rate);
			newwidth=boxW/rateint;
			newheight=boxH/rateint;
			}
		else
			{
			rateint=floor(rate);
			newwidth=boxW*rateint;
			newheight=boxH*rateint;
			}
			
		cropped=new unsigned int [boxW*boxH];
		k=0;
		for(int i=0;i<boxH;i++)
			{
			for(int j=0;j<boxW;j++)
				{
				cropped[k]=image[(upleftY+i)*height+upleftX+j];
				k++;
				}
			}
			
		//returnval=new char[newwidth * newheight];
		mytemplate=new char[20];
		srand(time(NULL));
		sprintf(mytemplate,"difftmp%dXXXXXX",rand()%100);
#ifdef _MSC_VER
		_mktemp(mytemplate);
#else
		mkstemp(mytemplate);
#endif
		mytempfile= fopen(mytemplate,"wt");
		test = fprintf(mytempfile,"P6 %d %d 255 ",newwidth,newheight);
        fclose(mytempfile);
					
		// validate the input
		if (image == NULL)
			throw (DiffractionImageException("NULL image"));
		
		try {
			firstcopy = new unsigned int[newwidth * newheight];
			copy1 = new unsigned char[newwidth * newheight];
			} catch (bad_alloc & e) {throw (DiffractionImageException("error allocating image copy"));}
		if (copy1 == NULL)
			throw (DiffractionImageException("error allocating image copy"));
		
		// I am not sure how these could fail, but wrap in a try/catch 
		// clause just to be sure.
		try	{
			//creating a larger or smaller image according to the ratio asked
			if(rate < 1)
				reduceImage(cropped,boxW,boxH,rateint,firstcopy);
			else
				enlargeImage(cropped,boxW,boxH,rateint,firstcopy);
			free(cropped);
			//changing the colouring scheme from to 256 level of gray.
			downSample(firstcopy, newwidth, newheight, true, copy1);
			free(firstcopy);
			invertImage(copy1, newwidth, newheight);
			} catch (DiffractionImageException e)
					{
					delete [] copy1;
					throw (e);
					}
		//preparing an array to write the ppm data.
		copy2 = new unsigned char[newwidth * newheight*3];
		for(int i=0;i<newwidth*newheight;i++)
			{
			copy2[i*3]=copy1[i];
			copy2[(i*3)+1]=copy1[i];
			copy2[(i*3)+2]=copy1[i];
			}
		free(copy1);
		filed = open(mytemplate,MYWRITE);
		lseek(filed,test,SEEK_SET);
		test = write(filed, copy2, newwidth*newheight*3);
		free(copy2);
        close(filed);
		return mytemplate;
		}
		
 }
