// DiffractionImageADSC.cpp
// maintained by G.Winter
// 17th December 2003
// 
// The bits of the DiffractionImage library which pertain to ADSC
// images.
// 
// 
// 
#include "DiffractionImage.h"
#include <string>
// if we are in development mode, then include things pertaining to assert.

using namespace std;

namespace DI 
	{

	// read in an ADSC image - this will read the header first, then 
	// reopen the file to read the unsigned shorts out - byte 
	// swapping if needed. NOTE WELL - this is hacky as there is 
	// an assumption here that the images will be represented as
	// unsigned shorts, which could well be wrong!
	void DiffractionImage::loadRAXISDtrek()
		{
		loadRAXISDtrekHeader();  // read the header
		allocateImage();
		loadRawImage(filename, originalHeaderLength);
		}
  
	void DiffractionImage::loadRAXISDtrekHeader()
		{
		float tempx, tempy;
		string temp,token;
		char* temp1;
		char* temp2;
		char* temp3;
		char* tempvalues;
		char littleString[2];
		int length, posDist, pos2T, count;
		//We only can be here if the loadADSC noticed it is a saturn image
		//Therefore we know that all keys/values have already been read in
		//we just need to populate the diffraction image values with the correct
		//keys.
		if (headerMap.find("header_bytes") == headerMap.end())
			throw(DiffractionImageException("header_bytes not found in header"));

		ASSERT_IN_HEADER("size1");
		ASSERT_IN_HEADER("size2");
		ASSERT_IN_HEADER("byte_order");
		ASSERT_IN_HEADER("scan_wavelength");
		ASSERT_IN_HEADER("rx_spatial_beam_position");
		ASSERT_IN_HEADER("rx_detector_size");
		ASSERT_IN_HEADER("rx_gonio_values");
		ASSERT_IN_HEADER("rx_gonio_names");
		ASSERT_IN_HEADER("rotation");

   		if (headerMap["byte_order"].compare("big_endian")==0)
			bigEndian = true;
		else
			bigEndian = false;
		// if we get here then everything we want is in the header
		// now fetch it all out into the header data structure
		// this will also be done via macro, with the following prototype
		GET_FROM_HEADER("size1", "%d", width);
		GET_FROM_HEADER("size2", "%d", height);
		GET_FROM_HEADER("scan_wavelength", "%f", wavelength);
		
		sscanf(headerMap["header_bytes"].c_str(), "%d",&originalHeaderLength);
		sscanf(headerMap["rx_detector_size"].c_str(), "%f %f", &tempx,&tempy);
		//NB, that is the beam position in pixels! not in mm
		sscanf(headerMap["rx_spatial_beam_position"].c_str(), "%f %f", &beamX, &beamY);
		//Therefore we need to do some calculations:
		pixelX=tempx/(width*1.0);
		pixelY=tempy/(height*1.0);
		beamX=beamX*pixelX;
		beamY=beamY*pixelY;
		exposureTime=0;
		sscanf(headerMap["rotation"].c_str(), "%f %f", &oscStart,&oscEnd);
		oscAxis=headerMap["rotation_axis_name"];
		length = headerMap["rx_gonio_names"].length();	
		littleString[1] = '\0';
		for(int i = 0; i < length; i++)
			{
			littleString[0] = (char) tolower(headerMap["rx_gonio_names"][i]);
			temp.append(littleString);
			}
		temp1=new char[length];
		strcpy(temp1,temp.c_str());
		
  		temp2 = strtok(temp1," ");
		count=0;
  		while (temp2 != NULL)
  			{
    		token=temp2;
			if(token.compare("distance") == 0 || token.compare("dist") == 0)
				posDist=count;
			if(token.compare("2theta") == 0 || token.compare("twotheta") == 0)
				pos2T=count;
    		temp2 = strtok(NULL, " ");
			count++;
  			}	
		length=headerMap["rx_gonio_values"].length();
		tempvalues=new char[length];
		strcpy(tempvalues,headerMap["rx_gonio_values"].c_str());
		temp3 = strtok(tempvalues," ");
		count=0;
  		while (temp3 != NULL)
  			{
			if(count==posDist)
				sscanf(temp3,"%f",&distance);
			if(count==pos2T)
				sscanf(temp3,"%f",&twoTheta);
    		temp3 = strtok(NULL, " ");
			count++;
			}
		// get the detector serial number out
		if (headerMap.find("rx_serial_number") != headerMap.end())
			serialNo = headerMap["rx_serial_number"];
		else
			serialNo="N/A";
		if (headerMap.find("rx_create_date") != headerMap.end())
			date=headerMap["rx_create_date"];
		else
			date="N/A";
		sscanf(headerMap["rx_detector_vectors"].c_str(), "%f %f %f %f %f %f",
		       &rapidAxis[0],&rapidAxis[1],&rapidAxis[2],
		       &slowAxis[0],&slowAxis[1],&slowAxis[2]);
		format="SMV";
		manufact="RIGAKU";
		}

	};
