// DiffractionImageUtility.cpp
// maintained by G.Winter
// 17th December 2003
// 
// Utility subroutines for the DiffractionImage module - things which 
// don't really belong anywhere else.
// 
// 
// 
// $Id: DiffractionImageUtility.cpp,v 1.11 2008/07/22 15:01:27 FR45 Exp $

#include "DiffractionImage.h"

using namespace std;

namespace DI
	{
    bool machineBigEndian(void)
		{
		unsigned short a = 0xff00;
		unsigned char * ptr = (unsigned char *) &a;
		if (ptr[0] == 0xff)
			return true; // I think that means big endian
		else
			return false;
		return false; // keep compiler san happy
		}

    // this could probably be done better inline - oh well!

    void byteSwap(unsigned short * blob)
		{
		unsigned char * ptr = (unsigned char *) blob;
		unsigned char temp;

		// I must be able to do this with masks and shifts
		temp = ptr[0];
		ptr[0] = ptr[1];
		ptr[1] = temp;
		}

    void byteSwap(unsigned int * blob)
		{
		unsigned char * ptr = (unsigned char *) blob;
		unsigned char temp;
		// I must be able to do this with masks and shifts

		temp = ptr[0];
		ptr[0] = ptr[3];
		ptr[3] = temp;
		temp = ptr[1];
		ptr[1] = ptr[2];
		ptr[2] = temp;
		}
		
	// try and guess the format of an image - based on the first two
    // bytes. adsc images begin {\n and marccd images begin 0x4949
    string guessFormat(string file)
		{
		int fd, amount;
		unsigned char buffer[4];
		fd = open(file.c_str(), O_RDONLY);
		if (fd < 0)
			{
			string msg="guessFormat had an error opening file: ";
			msg+=file;
			throw(DiffractionImageException(msg));
			}
		amount = read(fd, buffer, 4);
		if (amount != 4)
			{
			close(fd);
			string msg="guessFormat had an error reading file: ";
			msg+=file;
			throw(DiffractionImageException(msg));
			}
		close(fd);
		if ((buffer[0] == 0x49) && (buffer[1] == 0x49)) //this is a marccd image
			return string("mar");
		else
			{
			if (((buffer[0] == 0x0a) && (buffer[1] == 0x7b)) ||
				((buffer[0] == 0x7b) && (buffer[1] == 0x0a))) // this is an ADSC image
				return string("adsc");
			else
				{
				if (((buffer[0] == 'R') && (buffer[1] == 'A') && (buffer[2] == 'X') && (buffer[3] == 'I')) || 
					((buffer[0] == 'R') && (buffer[1] == '-') && (buffer[2] == 'A') && (buffer[3] == 'X')))
					return string("raxis");
				else
					{
					if ((buffer[0] == '#') && (buffer[1] == '#') && (buffer[2] == '#')) 
						return string("cbf"); 
					else
						{
						if ((buffer[0] == 'F') && (buffer[1] == 'O') && 
							(buffer[2] == 'R') && (buffer[3] == 'M')) 
							return string("bruker"); 
						else
							{
							if ((buffer[0] == 'O') && (buffer[1] == 'D') && 
								(buffer[2] == ' ')) 
								return string("oxford"); 
							else
								return string("unknown");
							}
						}
					}
				}
			}
		}
		
		// those other namespace sharing functions I was talking about

	// this function will take the three planar points (x0, y0) ...
	// (x2, y2) and fit a circle to them. clearly an identity between any 
	// two points will result in a degenerate circle - in this case the result
	// of the function will be 1. otherwise this should be successful, with 
	// return value 0 and the values r, x, y set to useful values.
	int fitCircle(float x0,float y0,float x1,float y1,float x2,float y2,float & x,float & y,float & r)
		{
		float a1, a2, b1, b2, c1, c2;

		// check for degeneracy - in the old code this used
		// x1 - x0 but this is a bad comparison function - change this to
		// fabs (x1 - x0) < epsilon, where the latter is a tiny number.

		// a useful point for scaling of the problem - these are coordinates
		// in pixels, so ought to be of order 1000. a suitable value for epsilon
		// will therefore be around 1.0
		const float epsilon = 1.0;
		if ((fabs(x0 - x1) < epsilon) || (fabs(x1 - x2) < epsilon) || (fabs(x2 - x0) < epsilon) ||
			(fabs(y0 - y1) < epsilon) || (fabs(y1 - y2) < epsilon) || (fabs(y2 - y0) < epsilon))
			return 1;
    
		// if we have reached this stage then we should be safe from degenerate points
		a1 = 2.0 * (x0 - x1);
		a2 = 2 * (x1 - x2);
		b1 = 2 * (y0 - y1);
		b2 = 2 * (y1 - y2);
		c1 = x0 * x0 - x1 * x1 + y0 * y0 - y1 * y1;
		c2 = x1 * x1 - x2 * x2 + y1 * y1 - y2 * y2;
		// get the centres

		x = (c1 * b2 - c2 * b1) / (a1 * b2 - a2 * b1);
		y = (c2 * a1 - c1 * a2) / (a1 * b2 - a2 * b1);
		
		// compute the radius from the first point - I am not sure that an
		// average would achieve anything here, other than highlight the 
		// rounding errors of floating point values
		r = sqrt((x0 - x) * (x0 - x) + (y0 - y) * (y0  - y));
		
		// all parameters computed, so job done
		return 0;
		}
		
	bool isInsideQuad(int* quad,int pointX,int pointY)
		{
		int* triangle;
		bool test;
		triangle=new int[6];
		triangle[0]=quad[0];
		triangle[1]=quad[1];
		triangle[2]=quad[2];
		triangle[3]=quad[3];
		triangle[4]=quad[4];
		triangle[5]=quad[5];
		test=isInsideTriangle(triangle,pointX,pointY);
		if(test)
			return true;
		else
			{
			triangle[0]=quad[0];
			triangle[1]=quad[1];
			triangle[2]=quad[4];
			triangle[3]=quad[5];
			triangle[4]=quad[6];
			triangle[5]=quad[7];
			test=isInsideTriangle(triangle,pointX,pointY);
			return test;
			}
		}
		
	bool isInsideTriangle(int *triangle,int pointX,int pointY)
		{
		int v1X,v1Y,v2X,v2Y,v3X,v3Y;
		//Those have to be double because integers might overflow!!
		double v1dotv1,v1dotv2,v1dotv3,v2dotv2,v2dotv3;
		double u,v;
		
		v1X=triangle[4]-triangle[0];	
		v2X=triangle[2]-triangle[0];
		v1Y=triangle[5]-triangle[1];	
		v2Y=triangle[3]-triangle[1];	
		v3X=pointX-triangle[0];
		v3Y=pointY-triangle[1];
			
		v1dotv1=v1X*v1X+v1Y*v1Y;
		v1dotv2=v1X*v2X+v1Y*v2Y;
		v1dotv3=v1X*v3X+v1Y*v3Y;
		v2dotv2=v2X*v2X+v2Y*v2Y;
		v2dotv3=v3X*v2X+v3Y*v2Y;
			
		u=(v2dotv2*v1dotv3-v1dotv2*v2dotv3)/((v1dotv1*v2dotv2)-(v1dotv2*v1dotv2));
		v=(v1dotv1*v2dotv3-v1dotv2*v1dotv3)/((v1dotv1*v2dotv2)-(v1dotv2*v1dotv2));
			
		return (((u>0 && v>0) && (u<1 && v<1)) && u+v<=1);
		}
	};
